package org.menacheri.jetserver.event.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorService;

import org.menacheri.jetserver.app.Session;
import org.menacheri.jetserver.concurrent.Command;
import org.menacheri.jetserver.concurrent.DataFlowVariable;
import org.menacheri.jetserver.concurrent.EventLoop;
import org.menacheri.jetserver.concurrent.Lane;
import org.menacheri.jetserver.event.Event;
import org.menacheri.jetserver.event.EventDispatcher;
import org.menacheri.jetserver.event.EventHandler;
import org.menacheri.jetserver.event.Events;
import org.menacheri.jetserver.event.SessionEventHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ABQEventDispatcher implements EventDispatcher
{
	private static final Logger LOG = LoggerFactory
			.getLogger(ABQEventDispatcher.class);
	private final EventLoop<ArrayBlockingQueue<Command>,ExecutorService> eventLoop;
	private final Lane<String, ExecutorService> dispatcherLane;
	private Map<Integer, List<EventHandler>> handlersByEventType;
	private List<EventHandler> anyHandler;
	private boolean isShuttingDown = false;

	public static class Entry implements Command
	{
		private ABQEventDispatcher eventDispatcher;
		private Event event;
		private Callable<Object> callable;
		private DataFlowVariable data;
		private Runnable runnable;

		public void execute()
		{
			if (null != event)
			{
				eventDispatcher.dispatchEvent(event);
			}
			else if (null != runnable)
			{
				runnable.run();
			}
			else if (null != callable && null != data)
			{
				try
				{
					data.bind(callable.call());
				}
				catch (Exception e)
				{
					throw new RuntimeException(e);
				}
			}
			else
			{
				throw new RuntimeException("Invalid " + this
						+ " object with all nulls put on queue");
			}
		}

		@Override
		public String toString()
		{
			return "Entry [eventDispatcher=" + eventDispatcher + ", event="
					+ event + ", callable=" + callable + ", data=" + data
					+ ", runnable=" + runnable + "]";
		}
	}

	public ABQEventDispatcher(EventLoop<ArrayBlockingQueue<Command>,ExecutorService> eventLoop,
			Lane<String, ExecutorService> lane)
	{
		this(eventLoop, lane, new HashMap<Integer, List<EventHandler>>(),
				new ArrayList<EventHandler>());
	}

	public ABQEventDispatcher(EventLoop<ArrayBlockingQueue<Command>,ExecutorService> eventLoop,
			Lane<String, ExecutorService> lane,
			Map<Integer, List<EventHandler>> handlersByEventType,
			List<EventHandler> genericHandlers)
	{
		this.eventLoop = eventLoop;
		this.dispatcherLane = lane;
		this.handlersByEventType = handlersByEventType;
		this.anyHandler = genericHandlers;
	}

	@Override
	public void addHandler(final EventHandler eventHandler)
	{
		Entry command = new Entry();
		command.runnable = new Runnable()
		{
			int eventType = eventHandler.getEventType();

			@Override
			public void run()
			{
				if (eventType == Events.ANY)
				{
					anyHandler.add(eventHandler);
				}
				else
				{
					List<EventHandler> handlers = handlersByEventType
							.get(eventType);
					if (handlers == null)
					{
						handlers = new CopyOnWriteArrayList<EventHandler>();
						handlersByEventType.put(eventType, handlers);
					}

					handlers.add(eventHandler);
				}
			}
		};
		eventLoop.publish(command);
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	@Override
	public List<EventHandler> getHandlers(final int eventType)
	{
		Entry command = new Entry();
		command.callable = new Callable<Object>()
		{
			@Override
			public Object call() throws Exception
			{
				return handlersByEventType.get(eventType);
			}
		};
		command.data = new DataFlowVariable();
		eventLoop.publish(command);
		try
		{
			return (List) command.data.getVal();
		}
		catch (InterruptedException e)
		{
			throw new RuntimeException(e);
		}
	}

	@Override
	public void removeHandler(final EventHandler eventHandler)
	{
		Entry command = new Entry();
		command.runnable = new Runnable()
		{

			@Override
			public void run()
			{
				int eventType = eventHandler.getEventType();
				if (eventType == Events.ANY)
				{
					anyHandler.remove(eventHandler);
				}
				else
				{
					List<EventHandler> handlers = handlersByEventType
							.get(eventType);
					if (null != handlers)
					{
						handlers.remove(eventHandler);
					}
				}
			}
		};
		eventLoop.publish(command);

	}

	@Override
	public void removeHandlersForEvent(final int eventType)
	{
		Entry command = new Entry();
		command.runnable = new Runnable()
		{

			@Override
			public void run()
			{
				List<EventHandler> handlers = handlersByEventType
						.get(eventType);
				if (null != handlers)
				{
					handlers.clear();
				}
			}
		};
		eventLoop.publish(command);
	}

	@Override
	public boolean removeHandlersForSession(Session session)
	{
		List<EventHandler> removeList = new ArrayList<EventHandler>();
		Collection<List<EventHandler>> eventHandlersList = handlersByEventType
				.values();
		for (List<EventHandler> handlerList : eventHandlersList)
		{
			if (null != handlerList)
			{
				for (EventHandler handler : handlerList)
				{
					if (handler instanceof SessionEventHandler)
					{
						SessionEventHandler sessionHandler = (SessionEventHandler) handler;
						if (sessionHandler.getSession().equals(session))
						{
							removeList.add(handler);
						}
					}
				}
			}
		}
		for (EventHandler handler : removeList)
		{
			removeHandler(handler);
		}
		return (removeList.size() > 0);
	}

	@Override
	public void fireEvent(Event event)
	{
		if (isShuttingDown)
		{
			LOG.warn(
					"Event: {} will be discarded since EventDispatcher is shutting down",
					event);
			return;
		}

		if (null != dispatcherLane
				&& dispatcherLane
						.isOnSameLane(Thread.currentThread().getName()))
		{
			dispatchEvent(event);
		}
		else
		{
			Entry command = new Entry();
			command.event = event;
			command.eventDispatcher = this;
			eventLoop.publish(command);
		}
	}

	protected void dispatchEvent(Event event)
	{
		for (EventHandler handler : anyHandler)
		{
			handler.onEvent(event);
		}

		// retrieval is not thread safe, but since we are not setting it to
		// null anywhere it should be fine.
		List<EventHandler> handlers = handlersByEventType.get(event.getType());
		// Iteration is thread safe since we use copy on write.
		if (null != handlers)
		{
			for (EventHandler handler : handlers)
			{
				handler.onEvent(event);
			}
		}
	}

	@Override
	public void close()
	{
		Entry command = new Entry();
		command.runnable = new Runnable()
		{
			@Override
			public void run()
			{
				isShuttingDown = true;
				anyHandler.clear();
				handlersByEventType.clear();
			}
		};
		eventLoop.publish(command);
	}

	@Override
	public void clear() {
		// TODO Auto-generated method stub
		
	}

}
